/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: ExtremePhysics.h
 * Include this file in your project to use the engine.
 */

#ifndef EXTREMEPHYSICS_H
#define EXTREMEPHYSICS_H

// ##### headers #####

#ifdef _MSC_VER
#define _USE_MATH_DEFINES // ExtremePhysics uses M_PI
#endif

#include <cfloat>
#include <cmath>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>

#ifdef _MSC_VER
#define EP_FORCEINLINE __forceinline
#else
#define EP_FORCEINLINE __attribute__((always_inline))
#endif

// ##### settings #####

// debugging
// You can disable debug checks in release mode but your program
// might crash if you do something wrong. A few examples:
// - using a polygon that has not been initialized
// - using a polygon that belongs to a different world
// - trying to connect a body with itself with a hinge joint
// - trying to connect bodies that belong to a different world
// - trying to move a ground body
// - ...
// Memory allocation will always be checked, even if you disable this.
#define EP_COMPILE_DEBUGCHECKS    1  // enable debug checks, eg. 'can not create hinge joint, both bodies are static'
#define EP_COMPILE_DEBUGMESSAGES  1  // compile debug messages
#define EP_COMPILE_DEBUGDRAW      1  // compile debug drawing

// 32-bit unsigned integer type
#define ep_int32   signed long
#define ep_uint32  unsigned long

// epsilon values used in floating-point calculations
#define EP_EPSILON_MINPOLYMASS       1.0e-20  // minimum mass (aera) for polygons
#define EP_EPSILON_MINPOLYINERTIA    1.0e-20  // minimum moment of inertia for polygons
#define EP_EPSILON_MINBODYMASS       1.0e-20  // minimum mass for bodies
#define EP_EPSILON_MINBODYINERTIA    1.0e-20  // minimum moment of inertia for bodies
#define EP_EPSILON_MINTIMESTEP       1.0e-20  // minimum length of a time step
#define EP_EPSILON_MAXCROSSPRODUCT   1.0e-4   // maximum cross product of adjacent edges (convexity check)
#define EP_EPSILON_MINPOLYEDGELEN    1.0e-8   // minimum length of polygon edges
#define EP_EPSILON_MINNORMALVECTOR   1.0e-8   // minimum length of a vector that will be normalized
#define EP_EPSILON_MINCONTACTPOINTS  1.0e-8   // minimum distance between two contact points
#define EP_EPSILON_MINCIRCLEVERTEX   1.0e-8   // minimum distance between the center of the circle and a vertex

// maximum number of free contacts that will be stored
#define EP_MAXFREECONTACTS 100

// size of the buffer used to create debug messages
#define EP_MESSAGE_BUFFERSIZE 200

// debug message levels
#define EP_MESSAGELEVEL_EXPLICIT         0  // this level of message is sent only when you explicitly ask for it
#define EP_MESSAGELEVEL_ERROR            1  // an error occurred
#define EP_MESSAGELEVEL_IMPORTANTCHANGE  2  // an important change occurred (create, destroy, ...)
#define EP_MESSAGELEVEL_NORMALCHANGE     3  // a normal change occurred (position changed, step, ...)

// shape types
#define EP_SHAPETYPE_BOX      1
#define EP_SHAPETYPE_CIRCLE   2
#define EP_SHAPETYPE_POLYGON  3

// constants
#define EP_PI M_PI
#define EP_MAXDOUBLE DBL_MAX // DBL_MIN = 1e-308, not -1e308! use -EP_MAXDOUBLE instead

// ##### macros #####

#define ep_sqr(x) ((x)*(x))
#define ep_abs(x) (fabs(x))
#define ep_max(a, b) (((a)>(b))? (a) : (b))
#define ep_min(a, b) (((a)<(b))? (a) : (b))
#define ep_clamp(a, b, val) (((val)<(a))? (a) : (((val)>(b))? (b) : (val)))
#define ep_round(x) (floor(0.5+(x)))

// rotates a vector
// [ x' ] = [  cos(a) ,  sin(a) ] * [ x ]
// [ y' ]   [ -sin(a) ,  cos(a) ]   [ y ]
#define ep_transform_x(sin, cos, x, y) (+(cos)*(x)+(sin)*(y))
#define ep_transform_y(sin, cos, x, y) (-(sin)*(x)+(cos)*(y))

// rotates a vector in the opposite direction
// cos(-a) = cos(a)
// sin(-a) = -sin(a)
#define ep_invtransform_x(sin, cos, x, y) (+(cos)*(x)-(sin)*(y))
#define ep_invtransform_y(sin, cos, x, y) (+(sin)*(x)+(cos)*(y))

// combines two transformations
// sin(a+b) = sin(a)*cos(b)+cos(a)*sin(b)
// cos(a+b) = cos(a)*cos(b)-sin(a)*sin(b)
#define ep_totaltransform_sin(sin1, cos1, sin2, cos2) ((sin1)*(cos2)+(cos1)*(sin2))
#define ep_totaltransform_cos(sin1, cos1, sin2, cos2) ((cos1)*(cos2)-(sin1)*(sin2))

// combines a transformation and an inversed transformation
// sin(a-b) = sin(a)*cos(b)-cos(a)*sin(b)
// cos(a-b) = cos(a)*cos(b)+sin(a)*sin(b)
#define ep_totalinvtransform_sin(sin1, cos1, sin2, cos2) ((sin1)*(cos2)-(cos1)*(sin2))
#define ep_totalinvtransform_cos(sin1, cos1, sin2, cos2) ((cos1)*(cos2)+(sin1)*(sin2))

// shuffles indices in an array
// count 5 -> n 5, {0,1,2,3,4} -> {1,3,0,2,4}
// count 6 -> n 7, {0,1,2,3,4,5} -> {1,3,5,0,2,4}
// count 7 -> n 7, {0,1,2,3,4,5,6} -> {1,3,5,0,2,4,6}
// count 8 -> n 9, {0,1,2,3,4,5,6,7} -> {1,3,5,7,0,2,4,6}
#define ep_shuffle_getn(count) ((count)|1)
#define ep_shuffle(iii, nnn) ((((iii)<<1)|1)%(nnn))

// ##### inline functions #####

// You can replace this with your allocator, but if memory allocation fails,
// it should either return NULL or throw an exception
inline void* ep_malloc(unsigned long size) { return malloc(size); }
inline void ep_free(void* mem) { free(mem); }

// ##### class prototypes #####

class ep_Callback;
class ep_Main;

class ep_World;

struct ep_PolygonVertex;
class ep_Polygon;

class ep_Group;
class ep_Body;

struct ep_CollisionData;
struct ep_ContactLink;
struct ep_ContactPoint;
class ep_Contact;

struct ep_HingeJointLink;
class ep_HingeJoint;

class ep_View;

class ep_Shape;

class ep_Force;

// ##### classes #####

// main

const char* ep_Version();

/// ExtremePhysics callback class.
/// Implement this class with your own functions.
/// Whatever you do, DON'T make ANY changes to ANY
/// ExtremePhysics object in the callback functions
/// or things will go wrong.
class ep_Callback {
	
	// public functions
	public:
	
	// Do NOT throw exceptions in this callback function! Check the return value of
	// the factory functions instead. Throwing exceptions inside this callback function
	// can cause large memory leaks in ep_World::UpdateContacts.
	virtual void DebugMessage(int level, const char* string) {}
	
	virtual void DrawBody(bool isstatic, bool issleeping, double x, double y) {}
	virtual void DrawSegment(double x1, double y1, double x2, double y2) {}
	virtual void DrawCircle(double x, double y, double r) {}
	
	virtual void DrawShapeLink(bool isstatic, double x1, double y1, double x2, double y2) {}
	virtual void DrawContactLink(bool isstatic, double x1, double y1, double x2, double y2) {}
	virtual void DrawJointLink(bool isstatic, double x1, double y1, double x2, double y2) {}
	
	virtual void DrawView(double x1, double y1, double x2, double y2) {}
	
	virtual void DrawVelocity(double x, double y, double xvel, double yvel, double rotvel) {}
	virtual void DrawForce(double x, double y, double xforce, double yforce, double torque) {}
	
	virtual void DrawContactPoint(double x1, double y1, double x2, double y2) {}
	virtual void DrawJointPoint(double x1, double y1, double x2, double y2) {}
	
};

/// The main ExtremePhysics class.
/// Create one instance of this class (on the stack or on the heap).
class ep_Main {
	
	friend class ep_World;
	friend class ep_Polygon;
	friend class ep_Body;
	friend class ep_Contact;
	friend class ep_HingeJoint;
	friend class ep_View;
	friend class ep_Shape;
	friend class ep_Force;
	
	// variables
	private:
	
	ep_Callback *callback;
	
	unsigned long idcounter_worlds;
	ep_World *first_world, *last_world, *current_world;
	
	int maxmessagelevel;
	
	// public functions
	public:
	
	/// Constructor
	/// @param _callback A pointer to an instance of ep_Callback (a derived class).
	ep_Main(ep_Callback* _callback = NULL);
	
	/// Destroys all ExtremePhysics objects in this world and resets id counters.
	void Clear();
	
	/// Creates a new world.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new world, or NULL if an error occurred.
	ep_World* CreateWorld(unsigned long userdatasize);
	/// Destroys a world.
	/// This function will also destroy all bodies, joints, ... associated with this world.
	/// @param world Pointer to the world to destroy.
	/// @return Returns whether successful.
	bool DestroyWorld(ep_World* world);
	/// Finds the world that corresponds to the given id.
	/// @param _id The id of the world.
	/// @return Returns a pointer to the world, or NULL if the world does not exist.
	ep_World* FindWorld(unsigned long _id);
	
	/// Shows a debug message. The format system is identical to printf().
	/// @param level The level of the message.
	/// @param format The format of the message.
	void Message(int level, const char* format, ...);
	/// Lists all ExtremePhysics objects (using debug messages).
	/// This is useful to detect memory leaks.
	void PrintObjectTree();
	
	// public inline functions
	public:
	
	/// Returns a pointer to the first world.
	inline ep_World* GetFirstWorld() { return first_world; }
	/// Returns a pointer to the last world.
	inline ep_World* GetLastWorld() { return last_world; }
	
	/// Sets the message filter.
	/// @param maxmessagelevel Messages with a level higher than this value will be ignored (performance).
	inline void SetMessageFilter(int _maxmessagelevel) { maxmessagelevel = _maxmessagelevel; }
	
};

// world

class ep_World {
	
	friend class ep_Main;
	friend class ep_Polygon;
	friend class ep_Body;
	friend class ep_Contact;
	friend class ep_HingeJoint;
	friend class ep_View;
	friend class ep_Shape;
	friend class ep_Force;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *prev, *next;
	unsigned long id;
	
	unsigned long idcounter_polygons, polygoncount;
	ep_Polygon *first_polygon, *last_polygon, *current_polygon;
	
	unsigned long idcounter_bodies, bodycount;
	ep_Body *first_body, *last_body, *current_body;
	
	unsigned long idcounter_contacts, contactcount;
	ep_Contact *first_contact, *last_contact, *current_contact;
	unsigned long freecontactcount;
	ep_Contact *freecontacts;
	
	unsigned long idcounter_hingejoints, hingejointcount;
	ep_HingeJoint *first_hingejoint, *last_hingejoint, *current_hingejoint;
	
	unsigned long idcounter_views, viewcount;
	ep_View *first_view, *last_view, *current_view;
	
	unsigned long shapecount;
	unsigned long forcecount;
	
	double timestep;
	unsigned long iterations;
	double contact_threshold, velocity_threshold, baumgarte_factor, mass_bias;
	
	bool horizontal;
	
	bool enable_sleeping;
	double time_stable, time_outofview;
	double stable_maxvel, stable_maxrotvel;
	
	// private functions
	private:
	
	void Init(ep_Main* _main);
	void DeInit();
	
	bool UpdateContacts();
	void UpdateSleeping();
	void IntegrateVelocity();
	void SolveVelocityConstraints();
	void IntegratePosition();
	
	ep_Contact* CreateContact(ep_Shape* shape1, ep_Shape* shape2);
	void DestroyContact(ep_Contact* contact);
	
	static bool Collision(ep_CollisionData* data, ep_Shape* shape1, ep_Shape* shape2, double contact_threshold);
	static bool CollisionBoxBox(ep_CollisionData* data, ep_Shape* shape1, ep_Shape* shape2, double contact_threshold);
	static bool CollisionBoxCircle(ep_CollisionData* data, ep_Shape* shape1, ep_Shape* shape2, double contact_threshold);
	static bool CollisionBoxPolygon(ep_CollisionData* data, ep_Shape* shape1, ep_Shape* shape2, double contact_threshold);
	static bool CollisionCircleCircle(ep_CollisionData* data, ep_Shape* shape1, ep_Shape* shape2, double contact_threshold);
	static bool CollisionCirclePolygon(ep_CollisionData* data, ep_Shape* shape1, ep_Shape* shape2, double contact_threshold);
	static bool CollisionPolygonPolygon(ep_CollisionData* data, ep_Shape* shape1, ep_Shape* shape2, double contact_threshold);
	
	void Awake();
	
	// public functions
	public:
	
	/// Destroys all objects in the world and resets id counters.
	void Clear();
	
	/// Finds the contact that corresponds to the given id.
	/// @param _id The id of the contact.
	/// @return Returns a pointer to the contact, or NULL if the contact does not exist.
	ep_Contact* FindContact(unsigned long _id);
	
	/// Creates a new polygon.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new polygon, or NULL if an error occurred.
	ep_Polygon* CreatePolygon(unsigned long vertexcount, unsigned long userdatasize);
	/// Destroys a polygon.
	/// Warning: This function will fail if the polygon is still in use.
	/// @param polygon Pointer to the polygon to destroy.
	/// @return Returns whether successful.
	bool DestroyPolygon(ep_Polygon* polygon);
	/// Finds the polygon that corresponds to the given id.
	/// @param _id The id of the polygon.
	/// @return Returns a pointer to the polygon, or NULL if the polygon does not exist.
	ep_Polygon* FindPolygon(unsigned long _id);
	
	/// Creates a new static body.
	/// @param groundbody Indicates whether the body is a ground body.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new body, or NULL if an error occurred.
	ep_Body* CreateStaticBody(bool groundbody, unsigned long userdatasize);
	/// Creates a new dynamic body.
	/// @param norotation Indicates whether the body can rotate.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new body, or NULL if an error occurred.
	ep_Body* CreateDynamicBody(bool norotation, unsigned long userdatasize);
	/// Destroys a body.
	/// This function will also destroy all shapes, forces and joints associated with this body.
	/// @param body Pointer to the body to destroy.
	/// @return Returns whether successful.
	bool DestroyBody(ep_Body* body);
	/// Finds the body that corresponds to the given id.
	/// @param _id The id of the body.
	/// @return Returns a pointer to the body, or NULL if the body does not exist.
	ep_Body* FindBody(unsigned long _id);
	
	/// Creates a new hinge joint.
	/// @param _body1 Pointer to the first body.
	/// @param _body2 Pointer to the second body.
	/// @param _x1 The x coordinate of the anchor point on the first body.
	/// @param _y1 The y coordinate of the anchor point on the first body.
	/// @param _x1 The x coordinate of the anchor point on the second body.
	/// @param _y1 The y coordinate of the anchor point on the second body.
	/// @param _referencerotation Reference used to calculate the relative rotation of the bodies (rot1-rot2).
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new hinge joint, or NULL if an error occurred.
	ep_HingeJoint* CreateHingeJoint(ep_Body* body1, ep_Body* body2, double x1, double y1, double x2, double y2, double referencerotation, unsigned long userdatasize);
	/// Destroys a hinge joint.
	/// @param body Pointer to the body to destroy.
	/// @return Returns whether successful.
	bool DestroyHingeJoint(ep_HingeJoint* hingejoint);
	/// Finds the hinge joint that corresponds to the given id.
	/// @param _id The id of thehinge joint.
	/// @return Returns a pointer to the hinge joint, or NULL if the hinge joint does not exist.
	ep_HingeJoint* FindHingeJoint(unsigned long _id);
	
	/// Creates a new view.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new view, or NULL if an error occurred.
	ep_View* CreateView(unsigned long userdatasize);
	/// Destroys a view.
	/// @param world Pointer to the world to destroy.
	/// @return Returns whether successful.
	bool DestroyView(ep_View* view);
	/// Finds the view that corresponds to the given id.
	/// @param id The id of the view.
	/// @return Returns a pointer to the view, or NULL if the world does not exist.
	ep_View* FindView(unsigned long _id);
	
	/// Changes the settings of the world.
	/// @param _timestep The length of one time step. The default value is 1.0.
	/// @param _iterations The number of iterations used by the velocity solver. Using a higher value increases accuracy. The default value is 20.
	/// @param _contact_threshold The threshold value for collisions. This stops contacts from being destroyed immediately after creation and increases stability.
	/// @param _velocity_threshold The velocity that is lost during collisions. This compensates for the energy collisions add to the system when gravity is used.
	/// @param _baumgarte_factor The bias factor used by the Baumgarte stabilization. Using a lower value will increase stability but decrease accuracy. The default value is 0.1.
	/// @param _mass_bias The bias factor used to calculate the mass of constraints. The best value depends on the complexity of the simulation. The default value is 0.6.
	void SetSettings(double _timestep, unsigned long _iterations, double _contact_threshold, double _velocity_threshold, double _baumgarte_factor, double _mass_bias);
	/// Changes the primary axis of the world.
	/// This doesn't change the results but the simulation will be faster if set correctly.
	/// Horizontal scanning is usually faster if there are lots of shapes with similar y-values.
	/// This is often the case when (vertical) gravity is used.
	/// @param _horizontal Indicates whether the world is mostly horizontal. The default value is true.
	void SetPrimaryAxis(bool _horizontal);
	/// Changes the sleeping settings of the world.
	/// @param _enable_sleeping Indicates whether sleeping should be enabled.
	/// @param _time_stable The time objects have to be stable before going to sleep. Set this to 0.0 to disable sleeping for stable objects.
	/// @param _time_outofview The time objects have to be out of view before going to sleep. Set this to 0.0 to disable sleeping for objects that are out of view.
	/// @param _stable_maxvel Bodies with a velocity lower than this value are considered stable.
	/// @param _stable_maxrotvel Bodies with a rotational velocity lowerthan this value are considered stable.
	void SetSleeping(bool _enable_sleeping, double _time_stable, double _time_outofview, double _stable_maxvel, double _stable_maxrotvel);
	/// Simulates one time step.
	/// @return Returns whether successful.
	bool SimulateStep();
	
	/// Draws all bodies in this world using the callback class.
	void DebugDrawBodies();
	/// Draws all pairs (contacts and joints) in this world using the callback class.
	void DebugDrawLinks();
	/// Draws all views in this world using the callback class.
	void DebugDrawViews();
	/// Draws the velocity of all bodies in this world using the callback class.
	void DebugDrawVelocity();
	/// Draws all forces (including contacts and joints) in this world using the callback class.
	void DebugDrawForces();
	/// Draws all points (contacts and joints) in this world using the callback class.
	void DebugDrawPoints();
	
	// public inline functions
	public:
	
	/// Returns the id of this world.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this world.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the previous world.
	inline ep_World* GetPrevious() { return prev; }
	/// Returns a pointer to the next world.
	inline ep_World* GetNext() { return next; }
	/// Returns a block to the userdata memory block.
	inline void* GetUserData() { return (void*)(this+1); }
	
	/// Returns a pointer to the first polygon.
	inline ep_Polygon* GetFirstPolygon() { return first_polygon; }
	/// Returns a pointer to the last polygon.
	inline ep_Polygon* GetLastPolygon() { return last_polygon; }
	/// Returns the number of polygons in this world.
	inline unsigned long GetPolygonCount() { return polygoncount; }
	
	/// Returns a pointer to the first body.
	inline ep_Body* GetFirstBody() { return first_body; }
	/// Returns a pointer to the last body.
	inline ep_Body* GetLastBody() { return last_body; }
	/// Returns the number of bodies in this world.
	inline unsigned long GetBodyCount() { return bodycount; }
	
	/// Returns a pointer to the first contact.
	inline ep_Contact* GetFirstContact() { return first_contact; }
	/// Returns a pointer to the last contact.
	inline ep_Contact* GetLastContact() { return last_contact; }
	/// Returns the number of contacts in this world.
	inline unsigned long GetContactCount() { return contactcount; }
	
	/// Returns a pointer to the first hinge joint.
	inline ep_HingeJoint* GetFirstHingeJoint() { return first_hingejoint; }
	/// Returns a pointer to the last hinge joint.
	inline ep_HingeJoint* GetLastHingeJoint() { return last_hingejoint; }
	/// Returns the number of hinge joints in this world.
	inline unsigned long GetHingeJointCount() { return hingejointcount; }
	
	/// Returns the number of shapes in this world.
	inline unsigned long GetShapeCount() { return shapecount; }
	/// Returns the number of forces in this world.
	inline unsigned long GetForceCount() { return forcecount; }
	
	/// Returns a pointer to the first view.
	inline ep_View* GetFirstView() { return first_view; }
	/// Returns a pointer to the last view.
	inline ep_View* GetLastView() { return last_view; }
	/// Returns the number of views in this world.
	inline unsigned long GetViewCount() { return viewcount; }
	
};

// polygon

struct ep_PolygonVertex {
	double x, y;
	double nx, ny, len, dist, pos;	
};
class ep_Polygon {
	
	friend class ep_Main;
	friend class ep_World;
	friend class ep_Body;
	friend class ep_Shape;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *world;
	ep_Polygon *prev, *next;
	unsigned long id;
	
	double xoff, yoff;
	double mass;
	double inertia;
	
	unsigned long vertexcount;
	ep_PolygonVertex *vertices;
	
#if EP_COMPILE_DEBUGCHECKS
	unsigned long referencecount;
	bool initialized;
#endif
	
	// private functions
	private:
	
	void Init(ep_World* _world, unsigned long _vertexcount);
	void DeInit();
	
	// public functions
	public:
	
	/// Sets the coordinates of a vertex of this polygon.
	/// @param index The index of the vertex, starting from 0.
	/// @param x The x coordinate of the vertex.
	/// @param y The y coordinate of the vertex.
	/// @return Returns whether successful.
	bool SetVertex(unsigned long index, double x, double y);
	/// Initializes this polygon.
	/// @return Returns whether successful.
	bool Initialize();
	
	// inline functions
	public:
	
	/// Returns the id of this polygon.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this polygon.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the world associated with this polygon.
	inline ep_World* GetWorld() { return world; }
	/// Returns a pointer to the previous polygon.
	inline ep_Polygon* GetPrevious() { return prev; }
	/// Returns a pointer to the next polygon.
	inline ep_Polygon* GetNext() { return next; }
	/// Returns a block to the userdata memory block.
	inline void* GetUserData() { return (void*)((ep_PolygonVertex*)(this+1)+vertexcount); }
	
};

// body

class ep_Body {
	
	friend class ep_Main;
	friend class ep_World;
	friend class ep_Contact;
	friend class ep_HingeJoint;
	friend class ep_Shape;
	friend class ep_Force;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *world;
	ep_Body *prev, *next;
	unsigned long id;
	
	unsigned long idcounter_shapes, shapecount;
	ep_Shape *first_shape, *last_shape, *current_shape;
	
	unsigned long idcounter_forces, forcecount;
	ep_Force *first_force, *last_force, *current_force;
	
	bool isgroundbody, isstatic, norotation;
	
	double xoff, yoff;
	double mass, invmass;
	double inertia, invinertia;
	
	double x, y, rot, rot_sin, rot_cos;
	double xvel, yvel, rotvel;
	
	double maxvel, maxrotvel;
	double gravity_x, gravity_y;
	double damping, rotdamping, damping_factor, rotdamping_factor;
	
	bool storecontactimpulses, storejointimpulses;
	bool sleepstable, sleepoutofview, issleeping;
	double stabletimer, outofviewtimer;
	
	ep_Body *nextislandbody;
	
	ep_ContactLink *first_contactlink, *last_contactlink;
	ep_HingeJointLink *first_hingejointlink, *last_hingejointlink;
	
	// private functions
	private:
	
	void Init(ep_World* _world, bool _isgroundbody, bool _isstatic, bool _norotation);
	void DeInit();
	
	void Awake(bool linkagechange, bool directchange);
	void Moved();
	
	EP_FORCEINLINE void _ApplyImpulse(double xx, double yy, double xforce, double yforce) {
		xvel += xforce*invmass;
		yvel += yforce*invmass;
		rotvel += (yy*xforce-xx*yforce)*invinertia;
	}
	EP_FORCEINLINE void _ApplyTorque(double torque) {
		rotvel += torque*invinertia;
	}
	
	// public functions
	public:
	
	/// Creates a new box shape.
	/// @param _w The width of the box.
	/// @param _h The height of the box.
	/// @param _x The x coordinate of the center of the shape relative to the body.
	/// @param _y The y coordinate of the center of the shape relative to the body.
	/// @param _rot The rotation of the shape relative to the body.
	/// @param _density The density of the shape.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new shape, or NULL if an error occurred.
	ep_Shape* CreateBoxShape(double _w, double _h, double _x, double _y, double _rot, double _density, unsigned long userdatasize);
	/// Creates a new circle shape.
	/// @param _r The radius of the circle.
	/// @param _x The x coordinate of the center of the shape relative to the body.
	/// @param _y The y coordinate of the center of the shape relative to the body.
	/// @param _rot The rotation of the shape relative to the body.
	/// @param _density The density of the shape.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new shape, or NULL if an error occurred.
	ep_Shape* CreateCircleShape(double _r, double _x, double _y, double _rot, double _density, unsigned long userdatasize);
	/// Creates a new polygon shape.
	/// @param _polygon Pointer to the polygon.
	/// @param _x The x coordinate of the center of the shape relative to the body.
	/// @param _y The y coordinate of the center of the shape relative to the body.
	/// @param _rot The rotation of the shape relative to the body.
	/// @param _density The density of the shape.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new shape, or NULL if an error occurred.
	ep_Shape* CreatePolygonShape(ep_Polygon* _polygon, double _x, double _y, double _rot, double _density, unsigned long userdatasize);
	/// Destroys a shape.
	/// @param world Pointer to the shape to destroy.
	/// @return Returns whether successful.
	bool DestroyShape(ep_Shape* shape);
	/// Finds the shape that corresponds to the given id.
	/// @param id The id of the shape.
	/// @return Returns a pointer to the shape, or NULL if the shape does not exist.
	ep_Shape* FindShape(unsigned long _id);
	
	/// Creates a new force.
	/// @param _x The x coordinate of the local point.
	/// @param _y The y coordinate of the local point.
	/// @param _local Indicates whether the force is local.
	/// @param _ignoremass Indicates whether the mass of the body should be ignored.
	/// @param userdatasize Size of the userdata memory block.
	/// @return Returns a pointer to the new world, or NULL if an error occurred.
	ep_Force* CreateForce(double _x, double _y, bool _local, bool _ignoremass, unsigned long userdatasize);
	/// Destroys a force.
	/// @param force Pointer to the force to destroy.
	/// @return Returns whether successful.
	bool DestroyForce(ep_Force* force);
	/// Finds the force that corresponds to the given id.
	/// @param id The id of the force.
	/// @return Returns a pointer to the force, or NULL if the force does not exist.
	ep_Force* FindForce(unsigned long _id);
	
	/// Returns a pointer to the first hinge joint connected to this body.
	/// @return Returns a pointer to the hinge joint, or NULL if no hinge joint is found.
	ep_HingeJoint* GetFirstHingeJoint();
	/// Returns a pointer to the last hinge joint connected to this body.
	/// @return Returns a pointer to the hinge joint, or NULL if no hinge joint is found.
	ep_HingeJoint* GetLastHingeJoint();
	/// Returns a pointer to the next hinge joint connected to this body.
	/// @param hingejoint Pointer to the current hinge joint.
	/// @return Returns a pointer to the hinge joint, or NULL if no hinge joint is found.
	ep_HingeJoint* GetPreviousHingeJoint(ep_HingeJoint* hingejoint);
	/// Returns a pointer to the previous hinge joint connected to this body.
	/// @param hingejoint Pointer to the current hinge joint.
	/// @return Returns a pointer to the hinge joint, or NULL if no hinge joint is found.
	ep_HingeJoint* GetNextHingeJoint(ep_HingeJoint* hingejoint);
	
	/// Calculates the mass, moment of inertia and center of mass of this body based on its shapes.
	/// @return Returns whether successful.
	bool CalculateMass();
	/// Changes the mass of the body.
	/// @param _mass The new mass.
	/// @return Returns whether successful.
	bool SetMass(double _mass);
	/// Changes the moment of inertia of the body.
	/// @param _mass The new moment of inertia.
	/// @return Returns whether successful.
	bool SetInertia(double _inertia);
	/// Changes the center of mass of the body.
	/// @param localx The x coordinate of the new center of mass.
	/// @param localy The y coordinate of the new center of mass.
	/// @param updateinertia Indicates whether to update the moment of inertia (using the parallel axis theorem).
	/// @return Returns whether successful.
	bool SetCenter(double localx, double localy, bool updateinertia);
	/// Changes the position of the body using the origin of the body as the reference.
	/// @param _x The new x coordinate of the body.
	/// @param _y The new y coordinate of the body.
	/// @param _rot The new rotation of the body.
	void SetPosition(double _x, double _y, double _rot);
	/// Changes the position of the body using the center of mass of the body as the reference.
	/// @param _x The new x coordinate of the body.
	/// @param _y The new y coordinate of the body.
	/// @param _rot The new rotation of the body.
	void SetPositionCenter(double _x, double _y, double _rot);
	/// Changes the position of the body using a local point of the body as the reference.
	/// @param _x The new x coordinate of the body.
	/// @param _y The new y coordinate of the body.
	/// @param _rot The new rotation of the body.
	/// @param localx The x coordinate of the reference point (local).
	/// @param localy The y coordinate of the reference point (local).
	void SetPositionLocalPoint(double _x, double _y, double _rot, double localx, double localy);
	/// Changes the velocity of the body using the center of mass of the body as the reference.
	/// @param _xvel The x component of the new velocity of the body.
	/// @param _yvel The y component of the new velocity of the body.
	/// @param _rotvel The new rotational velocity of the body.
	bool SetVelocityCenter(double _xvel, double _yvel, double _rotvel);
	/// Changes the velocity of the body using a local point of the body as the reference.
	/// @param _xvel The x component of the new velocity of the body.
	/// @param _yvel The new y velocity of the body.
	/// @param _rotvel The new rotational velocity of the body.
	/// @param localx The x coordinate of the reference point (local).
	/// @param localy The y coordinate of the reference point (local).
	bool SetVelocityLocalPoint(double _xvel, double _yvel, double _rotvel, double localx, double localy);
	
	/// Changes the maximum velocity of the body.
	/// @param _maxvel The new maximum velocity. Use 0.0 for no limit.
	/// @param _maxrotvel The new maximum rotational velocity. Use 0.0 for no limit.
	bool SetMaxVelocity(double _maxvel, double _maxrotvel);
	/// Changes the gravity of the body, ignoring the mass.
	/// @param _gravity_x The new x gravity of the body.
	/// @param _gravity_y The new y gravity of the body.
	bool SetGravity(double _gravity_x, double _gravity_y);
	/// Changes the damping factors of the body.
	/// @param _damping The damping factor of the body.
	/// @param _rotdamping The rotational damping factor of the body.
	bool SetDamping(double _damping, double _rotdamping);
	/// Changes stored impulses of the body.
	/// Impulses are stored only when this setting is set to true for both bodies.
	/// @param _storecontactimpulses Indicates whether contact impulses are stored. The default value is true.
	/// @param _storejointimpulses Indicates whether joint impulses are stored. The default value is false.
	void StoreImpulses(bool _storecontactimpulses, bool _storejointimpulses);
	/// Changes the sleeping settings of the body.
	/// These settings are ignored if sleeping when stable or sleeping out of view is disabled for the world.
	/// @param _sleepstable Indicates whether the body should sleep when it is stable. The default value is true.
	/// @param _sleepoutofview Indicates whether the body should sleep when it is out of view. The default value is true.
	bool SetSleeping(bool _sleepstable, bool _sleepoutofview);
	
	// public inline functions
	public:
	
	/// Returns the id of this body.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this body.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the world associated with this body.
	inline ep_World* GetWorld() { return world; }
	/// Returns a pointer to the previous body.
	inline ep_Body* GetPrevious() { return prev; }
	/// Returns a pointer to the next body.
	inline ep_Body* GetNext() { return next; }
	/// Returns a block to the userdata memory block.
	inline void* GetUserData() { return (void*)(this+1); }
	
	/// Returns a pointer to the first shape.
	inline ep_Shape* GetFirstShape() { return first_shape; }
	/// Returns a pointer to the last shape.
	inline ep_Shape* GetLastShape() { return last_shape; }
	/// Returns the number of shapes for this body.
	inline unsigned long GetShapeCount() { return shapecount; }
	
	/// Returns a pointer to the first polygon.
	inline ep_Force* GetFirstForce() { return first_force; }
	/// Returns a pointer to the last polygon.
	inline ep_Force* GetLastForce() { return last_force; }
	/// Returns the number of forces for this body.
	inline unsigned long GetForceCount() { return forcecount; }
	
	/// Returns the mass of the body.
	inline double GetMass() { return mass; }
	/// Returns the moment of inertia of the body.
	inline double GetInertia() { return inertia; }
	/// Returns the x coordinate of the center of mass of the body.
	inline double GetCenterOfMassX() { return xoff; }
	/// Returns the y coordinate of the center of mass of the body.
	inline double GetCenterOfMassY() { return yoff; }
	
	/// Returns the x coordinate of the body using the origin as the reference.
	inline double GetX() { return x-ep_transform_x(rot_sin, rot_cos, xoff, yoff); }
	/// Returns the y coordinate of the body using the origin as the reference.
	inline double GetY() { return y-ep_transform_y(rot_sin, rot_cos, xoff, yoff); }
	/// Returns the x coordinate of the body using the center of mass as the reference.
	inline double GetXCenter() { return x; }
	/// Returns the y coordinate of the body using the center of mass as the reference.
	inline double GetYCenter() { return y; }
	/// Returns the rotation of the body.
	inline double GetRot() { return rot; }
	
	/// Returns the x velocity of the body using the origin as the reference.
	inline double GetXVelCenter() { return xvel; }
	/// Returns the y velocity of the body using the origin as the reference.
	inline double GetYVelCenter() { return yvel; }
	/// Returns the x coordinate of the body using the origin as the reference.
	inline double GetXVelLocalPoint(double localx, double localy) { return xvel+ep_transform_y(rot_sin, rot_cos, localx-xoff, localy-yoff)*rotvel; }
	/// Returns the y coordinate of the body using the origin as the reference.
	inline double GetYVelLocalPoint(double localx, double localy) { return yvel-ep_transform_x(rot_sin, rot_cos, localx-xoff, localy-yoff)*rotvel; }
	/// Returns the rotational velocity of the body.
	inline double GetRotVel() { return yvel; }
	
	/// Returns whether the body is sleeping.
	inline bool IsSleeping() { return issleeping; }
	/// Returns the time this body has been stable.
	inline double StableTimer() { return stabletimer; }
	/// Returns the time this body has been out of view.
	inline double OutOfViewTimer() { return outofviewtimer; }
	
	/// Coordinate and vector conversions.
	inline double CoordLocalToWorldX(double localx, double localy) { return x+ep_transform_x(rot_sin, rot_cos, localx-xoff, localy-yoff); }
	inline double CoordLocalToWorldY(double localx, double localy) { return y+ep_transform_y(rot_sin, rot_cos, localx-xoff, localy-yoff); }
	inline double CoordWorldToLocalX(double worldx, double worldy) { return xoff+ep_invtransform_x(rot_sin, rot_cos, worldx-x, worldy-y); }
	inline double CoordWorldToLocalY(double worldx, double worldy) { return yoff+ep_invtransform_y(rot_sin, rot_cos, worldx-x, worldy-y); }
	inline double VectLocalToWorldX(double vx, double vy) { return ep_transform_x(rot_sin, rot_cos, vx, vy); }
	inline double VectLocalToWorldY(double vx, double vy) { return ep_transform_y(rot_sin, rot_cos, vx, vy); }
	inline double VectWorldToLocalX(double vx, double vy) { return ep_invtransform_x(rot_sin, rot_cos, vx, vy); }
	inline double VectWorldToLocalY(double vx, double vy) { return ep_invtransform_y(rot_sin, rot_cos, vx, vy); }
	
	/// Applies an impulse.
	/// @param localx The x coordinate of the local point.
	/// @param localy The y coordinate of the local point.
	/// @param xforce The x component of the force.
	/// @param yforce The y component of the force.
	/// @param torque The torque.
	/// @param local Indicates whether the force is local.
	/// @param ignoremass Indicates whether the mass and moment of inertia of the body should be ignored.
	/// @param awake Indicates whether the body should be awoken.
	inline void ApplyImpulse(double localx, double localy, double xforce, double yforce, double torque, bool local, bool ignoremass, bool awake) {
#if EP_COMPILE_DEBUGCHECKS
		if(isgroundbody) {
#if EP_COMPILE_DEBUGMESSAGES
			main->Message(EP_MESSAGELEVEL_ERROR, "Could not apply impulse to body %lu in world %lu, the body is a ground body.", id, world->id);
#endif
			return;
		}
#endif
		if(local) {
			xvel += ep_transform_x(rot_sin, rot_cos, xforce, yforce)*((ignoremass)? 1.0 : invmass);
			yvel += ep_transform_y(rot_sin, rot_cos, xforce, yforce)*((ignoremass)? 1.0 : invmass);
			rotvel +=
			  ((localy-yoff)*xforce-(localx-xoff)*yforce)*invinertia*((ignoremass)? mass : 1.0)
			  +torque*((ignoremass)? 1.0 : invinertia);
		} else {
			xvel += xforce*((ignoremass)? 1.0 : invmass);
			yvel += yforce*((ignoremass)? 1.0 : invmass);
			rotvel +=
			  ((localy-yoff)*ep_invtransform_x(rot_sin, rot_cos, xforce, yforce)
			  -(localx-xoff)*ep_invtransform_y(rot_sin, rot_cos, xforce, yforce))*invinertia*((ignoremass)? mass : 1.0)
			  +torque*((ignoremass)? 1.0 : invinertia);
		}
		if(awake) Awake(false, false);
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Applied impulse to body %lu in world %lu.", id, world->id);
#endif
	}
	
	/// Applies an impulse to a point relative to the center of mass.
	/// @param relativex The x coordinate relative to the center of mass.
	/// @param relativey The y coordinate relative to the center of mass.
	/// @param xforce The x component of the force.
	/// @param yforce The y component of the force.
	/// @param torque The torque.
	/// @param ignoremass Indicates whether the mass and moment of inertia of the body should be ignored.
	/// @param awake Indicates whether the body should be awoken.
	inline void ApplyImpulseRelative(double relativex, double relativey, double xforce, double yforce, double torque, bool ignoremass, bool awake) {
#if EP_COMPILE_DEBUGCHECKS
		if(isgroundbody) {
#if EP_COMPILE_DEBUGMESSAGES
			main->Message(EP_MESSAGELEVEL_ERROR, "Could not apply impulse to body %lu in world %lu, the body is a ground body.", id, world->id);
#endif
			return;
		}
#endif
		xvel += xforce*((ignoremass)? 1.0 : invmass);
		yvel += yforce*((ignoremass)? 1.0 : invmass);
		rotvel +=
		  (relativey*xforce-relativex*yforce)*invinertia*((ignoremass)? mass : 1.0)
		  +torque*((ignoremass)? 1.0 : invinertia);
		if(awake) Awake(false, false);
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Applied impulse to body %lu in world %lu.", id, world->id);
#endif
	}
	
};

// contact

struct ep_CollisionData {
	double nx, ny;
	bool cp_active[2];
	double cp_x[2][2], cp_y[2][2];
};

struct ep_ContactLink {
	ep_Contact *contact;
	ep_Shape *other_shape;
	ep_ContactLink *prev, *next;
};
struct ep_ContactPoint {
	bool active;
	double x1, y1, x2, y2;
	double xx1, yy1, xx2, yy2;
	double normaltargetvel, tangenttargetvel;
	double normalmass, tangentmass;
	double normalforce, tangentforce;
	double normalveldelta, tangentveldelta;
	/*
	double total_normalforce, total_tangentforce; //TODO// required? old substep system?
	double total_normalvel, total_tangentvel;
	*/
};
class ep_Contact {
	
	friend class ep_Main;
	friend class ep_World;
	friend class ep_Shape;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *world;
	ep_Contact *prev, *next;
	unsigned long id;
	
	ep_Body *body1, *body2;
	ep_Shape *shape1, *shape2;
	ep_ContactLink link1, link2;
	
	double nx, ny;
	double restitution, friction;
	
	ep_ContactPoint contactpoints[2];
	
	bool destroy;
	ep_Contact *nextconstraint;
	
	// private functions
	private:
	
	void Init(ep_World* _world, ep_Shape* _shape1, ep_Shape* _shape2);
	void DeInit();
	
	void Update(ep_CollisionData* data);
	void InitVelocityConstraints();
	void ApplyImpulses();
	void SolveVelocityConstraints();
	
	// public functions
	public:
	
	// inline functions
	public:
	
	/// Returns the id of this contact.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this contact.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the world associated with this contact.
	inline ep_World* GetWorld() { return world; }
	/// Returns a pointer to the previous contact.
	inline ep_Contact* GetPrevious() { return prev; }
	/// Returns a pointer to the next contact.
	inline ep_Contact* GetNext() { return next; }
	
	/// Returns a pointer to the first body associated with this contact.
	inline ep_Body* GetBody1() { return body1; }
	/// Returns a pointer to the second body associated with this contact.
	inline ep_Body* GetBody2() { return body2; }
	/// Returns a pointer to the first shape associated with this contact.
	inline ep_Shape* GetShape1() { return shape1; }
	/// Returns a pointer to the second shape associated with this contact.
	inline ep_Shape* GetShape2() { return shape2; }
	
	/// Returns the x component of the normal vector in world coordinates.
	inline double GetNormalX() { return nx; }
	/// Returns the y component of the normal vector in world coordinates.
	inline double GetNormalY() { return ny; }
	
	/// Returns whether the first contact point is active.
	inline bool GetPoint1Active() { return contactpoints[0].active; }
	/// Returns whether the second contact point is active.
	inline bool GetPoint2Active() { return contactpoints[1].active; }
	
	/// Returns the x coordinate of the first contact point in world coordinates.
	inline double GetPoint1X() {
		return (
		  body1->x+ep_transform_x(body1->rot_sin, body1->rot_cos, contactpoints[0].x1-body1->xoff, contactpoints[0].y1-body1->yoff)+
		  body2->x+ep_transform_x(body2->rot_sin, body2->rot_cos, contactpoints[0].x2-body2->xoff, contactpoints[0].y2-body2->yoff))/2.0;
	}
	/// Returns the y coordinate of the first contact point in world coordinates.
	inline double GetPoint1Y() {
		return (
		  body1->y+ep_transform_y(body1->rot_sin, body1->rot_cos, contactpoints[0].x1-body1->xoff, contactpoints[0].y1-body1->yoff)+
		  body2->y+ep_transform_y(body2->rot_sin, body2->rot_cos, contactpoints[0].x2-body2->xoff, contactpoints[0].y2-body2->yoff))/2.0;
	}
	/// Returns the x coordinate of the second contact point in world coordinates.
	inline double GetPoint2X() {
		return (
		  body1->x+ep_transform_x(body1->rot_sin, body1->rot_cos, contactpoints[1].x1-body1->xoff, contactpoints[1].y1-body1->yoff)+
		  body2->x+ep_transform_x(body2->rot_sin, body2->rot_cos, contactpoints[1].x2-body2->xoff, contactpoints[1].y2-body2->yoff))/2.0;
	}
	/// Returns the y coordinate of the second contact point in world coordinates.
	inline double GetPoint2Y() {
		return (
		  body1->y+ep_transform_y(body1->rot_sin, body1->rot_cos, contactpoints[1].x1-body1->xoff, contactpoints[1].y1-body1->yoff)+
		  body2->y+ep_transform_y(body2->rot_sin, body2->rot_cos, contactpoints[1].x2-body2->xoff, contactpoints[1].y2-body2->yoff))/2.0;
	}
	
	/// Returns the normal force of the first contact point.
	inline double GetPoint1NormalForce() { return contactpoints[0].normalforce; }
	/// Returns the tangent force of the first contact point.
	inline double GetPoint1TangentForce() { return contactpoints[0].tangentforce; }
	/// Returns the normal force of the second contact point.
	inline double GetPoint2NormalForce() { return contactpoints[1].normalforce; }
	/// Returns the tangent force of the second contact point.
	inline double GetPoint2TangentForce() { return contactpoints[1].tangentforce; }
	
	/// Returns the normal velocity difference of the first contact point.
	inline double GetPoint1NormalVelDelta() { return contactpoints[0].normalveldelta; }
	/// Returns the tangent velocity difference of the first contact point.
	inline double GetPoint1TangentVelDelta() { return contactpoints[0].tangentveldelta; }
	/// Returns the normal velocity difference of the second contact point.
	inline double GetPoint2NormalVelDelta() { return contactpoints[1].normalveldelta; }
	/// Returns the tangent velocity difference of the second contact point.
	inline double GetPoint2TangentVelDelta() { return contactpoints[1].tangentveldelta; }
	
};

// hinge joint

struct ep_HingeJointLink {
	ep_HingeJoint *hingejoint;
	ep_Body *other_body;
	ep_HingeJointLink *prev, *next;
};
class ep_HingeJoint {
	
	friend class ep_Main;
	friend class ep_World;
	friend class ep_Body;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *world;
	ep_HingeJoint *prev, *next;
	unsigned long id;
	
	ep_Body *body1, *body2;
	ep_HingeJointLink link1, link2;
	
	double x1, y1, x2, y2;
	double referencerotation, maxforce;
	
	double maxmotortorque, motorvel;
	
	double limit1_maxtorque, limit2_maxtorque;
	double limit1_rot, limit2_rot;
	
	double xx1, yy1, xx2, yy2, targetvel_x, targetvel_y;
	double targetlimitvel, minlimittorque, maxlimittorque;
	double m1_x, m1_y, m2_x, m2_y, motormass;
	
	double xforce, yforce, torque;
	double limittorque;
	
	ep_HingeJoint *nextconstraint;
	
	// private functions
	private:
	
	void Init(ep_World* _world, ep_Body* _body1, ep_Body* _body2, double _x1, double _y1, double _x2, double _y2, double _referencerotation);
	void DeInit();
	
	void InitVelocityConstraints();
	void ApplyImpulses();
	void SolveVelocityConstraints();
	
	// public functions
	public:
	
	/// Sets the maximum force of this hinge joint
	/// @param _maxforce The new maximum force. Use 0.0 for no limit.
	void SetMaxForce(double _maxforce);
	/// Sets the motor torque and velocity of this hinge joint
	/// @param _maxmotortorque The new maximum motor torque. Use 0.0 to disable the motor.
	/// @param _motorvel The new motor velocity.
	void SetMotor(double _maxmotortorque, double _motorvel);
	/// Sets the lower limit of the hinge joint.
	/// @param maxlimittorque The maximum limit torque. Use 0.0 to disable the limit.
	/// @param rotation The lower limit of the rotation.
	void SetLowerLimit(double maxlimittorque, double rotation);
	/// Sets the upper limit of the hinge joint.
	/// @param maxlimittorque The maximum limit torque. Use 0.0 to disable the limit.
	/// @param rotation The upper limit of the rotation.
	void SetUpperLimit(double maxlimittorque, double rotation);
	
	// inline functions
	public:
	
	/// Returns the id of this hinge joint.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this hinge joint.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the world associated with this hinge joint.
	inline ep_World* GetWorld() { return world; }
	/// Returns a pointer to the previous hinge joint.
	inline ep_HingeJoint* GetPrevious() { return prev; }
	/// Returns a pointer to the next hinge joint.
	inline ep_HingeJoint* GetNext() { return next; }
	/// Returns a block to the userdata memory block.
	inline void* GetUserData() { return (void*)(this+1); }
	
	/// Returns a pointer to the first body associated with this hinge joint.
	inline ep_Body* GetBody1() { return body1; }
	/// Returns a pointer to the second body associated with this hinge joint.
	inline ep_Body* GetBody2() { return body2; }
	
	/// Returns the current rotation of the hinge joint.
	inline double GetRotation() { return body1->rot-body2->rot-referencerotation; };
	/// Returns the x coordinate of the force applied during the last step.
	inline double GetXForce() { return xforce; };
	/// Returns the y coordinate of the force applied during the last step.
	inline double GetYForce() { return yforce; };
	/// Returns the torque applied during the last step.
	inline double GetTorque() { return torque; };
	
};

// view

class ep_View {
	
	friend class ep_Main;
	friend class ep_World;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *world;
	ep_View *prev, *next;
	unsigned long id;
	
	double x1, y1, x2, y2;
	
	// private functions
	private:
	
	void Init(ep_World* _world);
	void DeInit();
	
	// public functions
	public:
	
	/// Changes the rectangle of the view.
	/// @param _x1 The left side of the view rectangle.
	/// @param _y1 The top side of the view rectangle.
	/// @param _x2 The right side of the view rectangle.
	/// @param _y2 The bottom side of the view rectangle.
	void SetRectangle(double _x1, double _y1, double _x2, double _y2);
	
	// inline functions
	public:
	
	/// Returns the id of this view.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this view.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the world associated with this view.
	inline ep_World* GetWorld() { return world; }
	/// Returns a pointer to the previous view.
	inline ep_View* GetPrevious() { return prev; }
	/// Returns a pointer to the next view.
	inline ep_View* GetNext() { return next; }
	/// Returns a block to the userdata memory block.
	inline void* GetUserData() { return (void*)(this+1); }
	
};

// shape

class ep_Shape {
	
	friend class ep_Main;
	friend class ep_World;
	friend class ep_Body;
	friend class ep_Contact;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *world;
	ep_Body *body;
	ep_Shape *prev, *next;
	unsigned long id;
	
	int shapetype;
	union {
		struct {
			double w, h;
		} box;
		struct {
			double r;
		} circle;
		struct {
			ep_Polygon *polygon;
		} polygon;
	} shapedata;
	
	double x, y, rot_sin, rot_cos, density;
	
	float restitution, friction;
	double normalvelocity, tangentvelocity;
	unsigned long collidemask1, collidemask2, group;
	
	bool updateaabb, updatecontacts;
	double aabb_x1, aabb_y1, aabb_x2, aabb_y2;
	ep_uint32 aabbul_x1, aabbul_y1, aabbul_x2, aabbul_y2;
	double xxx, yyy, sin_t, cos_t;
	ep_Contact *temp_c;
	
	ep_ContactLink *first_contactlink, *last_contactlink;
	
	// private functions
	private:
	
	void Init(ep_Body* _body, int _shapetype, double _x, double _y, double _rot, double _density, double arg1, double arg2, ep_Polygon* arg3);
	void DeInit();
	
	void UpdateAABB();
	
	// public functions
	public:
	
	/// Returns a pointer to the first contact connected to this shape.
	/// @return Returns a pointer to the contact, or NULL if no contact is found.
	ep_Contact* GetFirstContact();
	/// Returns a pointer to the last contact connected to this shape.
	/// @return Returns a pointer to the contact, or NULL if no contact is found.
	ep_Contact* GetLastContact();
	/// Returns a pointer to the next contact connected to this shape.
	/// @param hingejoint Pointer to the current contact.
	/// @return Returns a pointer to the contact, or NULL if no contact is found.
	ep_Contact* GetPreviousContact(ep_Contact* contact);
	/// Returns a pointer to the previous contact connected to this shape.
	/// @param hingejoint Pointer to the current contact.
	/// @return Returns a pointer to the contact, or NULL if no contact is found.
	ep_Contact* GetNextContact(ep_Contact* contact);
	
	/// Changes the material of the shape.
	/// @param _restitution The coefficient of restitution.
	/// @param _friction The coefficient of friction.
	/// @param _normalvelocity The normal velocity of the surface.
	/// @param _tangentvelocity The tangent velocity of the surface.
	void SetMaterial(float _restitution, float _friction, double _normalvelocity, double _tangentvelocity);
	/// Changes the collision settings of the shape.
	/// @param _collidemask1 The first collision mask.
	/// @param _collidemask2 The second collision mask.
	/// @param _group The group of the shape. Use zero for no group.
	void SetCollision(unsigned long _collidemask1, unsigned long _collidemask2, unsigned long _group);
	
	// inline functions
	public:
	
	/// Returns the id of this shape.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this shape.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the world associated with this shape.
	inline ep_World* GetWorld() { return world; }
	/// Returns a pointer to the body associated with this shape.
	inline ep_Body* GetBody() { return body; }
	/// Returns a pointer to the previous shape.
	inline ep_Shape* GetPrevious() { return prev; }
	/// Returns a pointer to the next shape.
	inline ep_Shape* GetNext() { return next; }
	/// Returns a block to the userdata memory block.
	inline void* GetUserData() { return (void*)(this+1); }
	
};

// force

class ep_Force {
	
	friend class ep_Main;
	friend class ep_World;
	friend class ep_Body;
	
	// variables
	private:
	
	ep_Main *main;
	ep_World *world;
	ep_Body *body;
	ep_Force *prev, *next;
	unsigned long id;
	
	double x, y;
	double xforce, yforce, torque;
	bool local, ignoremass;
	
	// private functions
	private:
	
	void Init(ep_Body* _body, double _x, double _y, bool _local, bool _ignoremass);
	void DeInit();
	
	// inline functions
	public:
	
	/// Returns the id of this force.
	inline unsigned long GetID() { return id; }
	/// Returns a pointer to the main ExtremePhysics object associated with this force.
	inline ep_Main* GetMain() { return main; }
	/// Returns a pointer to the world associated with this force.
	inline ep_World* GetWorld() { return world; }
	/// Returns a pointer to the body associated with this force.
	inline ep_Body* GetBody() { return body; }
	/// Returns a pointer to the previous force.
	inline ep_Force* GetPrevious() { return prev; }
	/// Returns a pointer to the next force.
	inline ep_Force* GetNext() { return next; }
	/// Returns a block to the userdata memory block.
	inline void* GetUserData() { return (void*)(this+1); }
	
	/// Changes the force.
	/// @param _xforce The x component of the new force.
	/// @param _yforce The y component of the new force.
	/// @param _torque The new torque.
	inline void SetForce(double _xforce, double _yforce, double _torque, double awake) {
		if(xforce!=_xforce || yforce!=_yforce || torque!=_torque) {
			xforce = _xforce;
			yforce = _yforce;
			torque = _torque;
			if(awake) body->Awake(false, false);
		}
#if EP_COMPILE_DEBUGMESSAGES
		main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed force of force %lu in world %lu.", id, world->id);
#endif
	}
	
};

#endif // EXTREMEPHYSICS_H

