/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: ep_HingeJoint_VelocityConstraints.h
 * Solves hinge joint velocity constraints.
 * Included in ep_World_SolveVelocityConstraints.cpp.
 */

#include "ExtremePhysics.h"

void ep_HingeJoint::InitVelocityConstraints() {
	
	// reset impulses
	/*
	//TODO// remove
	if(!body1->storejointimpulses || !body2->storejointimpulses) {
		xforce = 0.0;
		yforce = 0.0;
		torque = 0.0;
	}
	*/
	
	// update relative coordinates
	xx1 = ep_transform_x(body1->rot_sin, body1->rot_cos, x1-body1->xoff, y1-body1->yoff);
	yy1 = ep_transform_y(body1->rot_sin, body1->rot_cos, x1-body1->xoff, y1-body1->yoff);
	xx2 = ep_transform_x(body2->rot_sin, body2->rot_cos, x2-body2->xoff, y2-body2->yoff);
	yy2 = ep_transform_y(body2->rot_sin, body2->rot_cos, x2-body2->xoff, y2-body2->yoff);
	
	// baumgarte
	targetvel_x = -world->baumgarte_factor*((body1->x+xx1)-(body2->x+xx2));
	targetvel_y = -world->baumgarte_factor*((body1->y+yy1)-(body2->y+yy2));
	
	// limits
	double rot = body1->rot-body2->rot-referencerotation;
	if(limit1_maxtorque!=0.0 && limit2_maxtorque!=0.0 && limit1_rot==limit2_rot) {
		targetlimitvel = -world->baumgarte_factor*(rot-limit1_rot);
		maxlimittorque = +limit1_maxtorque;
		minlimittorque = -limit2_maxtorque;
	} else {
		if(limit1_maxtorque!=0.0 && rot<limit1_rot) {
			maxlimittorque = +limit1_maxtorque;
			targetlimitvel = -world->baumgarte_factor*(rot-limit1_rot);
		} else {
			maxlimittorque = 0.0;
		}
		if(limit2_maxtorque!=0.0 && rot>limit2_rot) {
			minlimittorque = -limit2_maxtorque;
			targetlimitvel = -world->baumgarte_factor*(rot-limit2_rot);
		} else {
			minlimittorque = 0.0;
		}
	}
	
	{
		// set up matrix
		double aa, bb, cc, dd;
		aa = body1->invmass
			+body2->invmass
			+body1->invinertia*yy1*yy1
			+body2->invinertia*yy2*yy2;
		bb = 0.0
			+0.0
			-body1->invinertia*yy1*xx1
			-body2->invinertia*yy2*xx2;
		cc = 0.0
			+0.0
			-body1->invinertia*xx1*yy1
			-body2->invinertia*xx2*yy2;
		dd = body1->invmass
			+body2->invmass
			+body1->invinertia*xx1*xx1
			+body2->invinertia*xx2*xx2;
		
		// invert (create mass matrix)
		double d = (1.0+world->mass_bias)/(aa*dd-bb*cc);
		m1_x = d*dd;
		m2_x = -d*bb;
		m1_y = -d*cc;
		m2_y = d*aa;
		
	}
	
	// calculate motor/limit mass
	if(maxmotortorque!=0.0 || minlimittorque!=maxlimittorque) {
		motormass = 1.0/(body1->invinertia+body2->invinertia);
	}
	
}

void ep_HingeJoint::ApplyImpulses() {
	
	if(body1->storejointimpulses && body2->storejointimpulses) {
		
		// clamp stored impulse (maybe maxforce was changed)
		if(maxforce!=0.0) {
			double d = ep_sqr(xforce)+ep_sqr(yforce);
			if(d>ep_sqr(maxforce)) {
				d = sqrt(d);
				xforce *= maxforce/d;
				yforce *= maxforce/d;
			}
		}
		
		// clamp stored torque
		torque = ep_clamp(-maxmotortorque, +maxmotortorque, torque);
		limittorque = ep_clamp(minlimittorque,  maxlimittorque, limittorque);
		
		body1->_ApplyImpulse(xx1, yy1, +xforce, +yforce);
		body2->_ApplyImpulse(xx2, yy2, -xforce, -yforce);
		body1->_ApplyTorque(+torque+limittorque);
		body2->_ApplyTorque(-torque-limittorque);
		
	} else {
		xforce = 0.0;
		yforce = 0.0;
		torque = 0.0;
		limittorque = 0.0;
	}
	
}

EP_FORCEINLINE void ep_HingeJoint::SolveVelocityConstraints() {
	
	/*
	// motor
	if(maxmotortorque!=0.0) {
		
		double currrotvel = body1->rotvel-body2->rotvel;
		double f = (motorvel-currrotvel)*motormass;
		
		// always clamp the accumulated impulse
		if(torque+f<-maxmotortorque) {
			f = -maxmotortorque-torque;
			torque = -maxmotortorque;
		} else if(torque+f>maxmotortorque) {
			f = +maxmotortorque-torque;
			torque = +maxmotortorque;
		} else {
			torque += f;
		}
		
		if(f!=0.0) {
			body1->_ApplyTorque(+f);
			body2->_ApplyTorque(-f);
		}
		
	}
	
	// limits
	// minlimittorque<=0 and maxlimittorque>=0
	if(minlimittorque!=maxlimittorque) {
		
		double currrotvel = body1->rotvel-body2->rotvel;
		double f = (targetlimitvel-currrotvel)*motormass;
		
		// always clamp the accumulated impulse
		if(limittorque+f<-minlimittorque) {
			f = minlimittorque-limittorque;
			limittorque = minlimittorque;
		} else if(limittorque+f>maxlimittorque) {
			f = maxlimittorque-limittorque;
			limittorque = maxlimittorque;
		} else {
			limittorque += f;
		}
		
		if(f!=0.0) {
			body1->_ApplyTorque(+f);
			body2->_ApplyTorque(-f);
		}
		
	}
	*/
	
	// motor/limits
	if(maxmotortorque!=0.0 || minlimittorque!=maxlimittorque) {
		
		double currrotvel = body1->rotvel-body2->rotvel;
		double combinedtorque;
		
		if(maxmotortorque!=0.0) {
			
			combinedtorque = (motorvel-currrotvel)*motormass;
			
			// always clamp the accumulated impulse
			if(torque+combinedtorque<-maxmotortorque) {
				combinedtorque = -maxmotortorque-torque;
				torque = -maxmotortorque;
			} else if(torque+combinedtorque>maxmotortorque) {
				combinedtorque = +maxmotortorque-torque;
				torque = +maxmotortorque;
			} else {
				torque += combinedtorque;
			}
			
			currrotvel += combinedtorque/motormass;
			
		} else {
			combinedtorque = 0.0;
		}
		
		if(minlimittorque!=maxlimittorque) {
			
			double f = (targetlimitvel-currrotvel)*motormass;
			
			// always clamp the accumulated impulse
			if(limittorque+f<minlimittorque) {
				f = minlimittorque-limittorque;
				limittorque = minlimittorque;
			} else if(limittorque+f>maxlimittorque) {
				f = maxlimittorque-limittorque;
				limittorque = maxlimittorque;
			} else {
				limittorque += f;
			}
			
			combinedtorque += f;
			
		}
		
		if(combinedtorque!=0.0) {
			body1->_ApplyTorque(+combinedtorque);
			body2->_ApplyTorque(-combinedtorque);
		}
		
	}
	
	double xx = targetvel_x-((body1->xvel+yy1*body1->rotvel)-(body2->xvel+yy2*body2->rotvel)),
	       yy = targetvel_y-((body1->yvel-xx1*body1->rotvel)-(body2->yvel-xx2*body2->rotvel));
	double fx = xx*m1_x+yy*m2_x,
	       fy = xx*m1_y+yy*m2_y;
	
	if(maxforce==0.0) {
		xforce += fx;
		yforce += fy;
	} else {
		double xxx = xforce+xx;
		double yyy = yforce+yy;
		double d = ep_sqr(xxx)+ep_sqr(yyy);
		if(d>ep_sqr(maxforce)) {
			d = sqrt(d);
			xxx *= maxforce/d;
			yyy *= maxforce/d;
			fx = xxx-xforce;
			fy = yyy-yforce;
			xforce = xxx;
			yforce = yyy;
		} else {
			xforce += fx;
			yforce += fy;
		}
	}
	
	body1->_ApplyImpulse(xx1, yy1, +fx, +fy);
	body2->_ApplyImpulse(xx2, yy2, -fx, -fy);
	
}

/*
Hinge joint constraint derivation notes:

constraint: relative speed at joint anchor (#) is always zero
(# = relative coordinates of joint anchor)

delta_xspeed = fx/mass1+fx/mass2+( yy1#)*(fx*yy1-fy*xx1)/inertia1+( yy2#)*(fx*yy2-fy*xx2)/inertia2;
delta_yspeed = fy/mass1+fy/mass2+(-xx1#)*(fx*yy1-fy*xx1)/inertia1+(-xx2#)*(fx*yy2-fy*xx2)/inertia2;

delta_xspeed = 
fx*(1/mass1)                + 0                           +
fx*(1/mass2)                + 0                           +
fx*yy1#*yy1*(1/inertia1)    + -fy*yy1#*xx1*(1/inertia1)   +
/fx*yy2#*yy2*(1/inertia2)    + -fy*yy2#*xx2*(1/inertia2)   ;

delta_yspeed = 
0                           + fy*(1/mass1)                +
0                           + fy*(1/mass2)                +
-fx*xx1#*yy1*(1/inertia1)   + fy*xx1#*xx1*(1/inertia1)    +
-fx*xx2#*yy2*(1/inertia2)   + fy*xx2#*xx2*(1/inertia2)    ;

matrix:
  [ (1/mass1) , 0         ] + [ (1/mass2) , 0         ]
  [ 0         , (1/mass1) ]   [ 0         , (1/mass2) ]
+ [  yy1#*yy1*(1/inertia1) , -yy1#*xx1*(1/inertia1) ]
  [ -xx1#*yy1*(1/inertia1) ,  xx1#*xx1*(1/inertia1) ]
+ [  yy2#*yy2*(1/inertia2) , -yy2#*xx2*(1/inertia2) ]
  [ -xx2#*yy2*(1/inertia2) ,  xx2#*xx2*(1/inertia2) ]
the determinant is always positive if 1/mass1+1/mass2>=0
*/

