/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: ep_Shape.cpp
 * Implementation of ep_Shape.
 */

#include "ExtremePhysics.h"

void ep_Shape::Init(ep_Body* _body, int _shapetype, double _x, double _y, double _rot, double _density, double arg1, double arg2, ep_Polygon* arg3) {
	
	// add to linked list
	main = _body->main;
	world = _body->world;
	body = _body;
	prev = body->last_shape;
	next = NULL;
	if(prev==NULL) body->first_shape = this;
	else prev->next = this;
	body->last_shape = this;
	body->current_shape = this;
	
	// initialize variables
	id = ++body->idcounter_shapes;
	
	shapetype = _shapetype;
	switch(shapetype) {
		case EP_SHAPETYPE_BOX: {
			shapedata.box.w = arg1;
			shapedata.box.h = arg2;
			break;
		}
		case EP_SHAPETYPE_CIRCLE: {
			shapedata.circle.r = arg1;
			break;
		}
		case EP_SHAPETYPE_POLYGON: {
			shapedata.polygon.polygon = arg3;
			break;
		}
	}
	
	x = _x;
	y = _y;
	rot_sin = sin(_rot);
	rot_cos = cos(_rot);
	density = _density;
	
	restitution = 0.0;
	friction = 0.0;
	normalvelocity = 0.0;
	tangentvelocity = 0.0;
	collidemask1 = 0;
	collidemask2 = 0;
	group = 0;
	
	updateaabb = true;
	updatecontacts = true;
	
	first_contactlink = NULL;
	last_contactlink = NULL;
	
#if EP_COMPILE_DEBUGCHECKS
	if(shapetype==EP_SHAPETYPE_POLYGON) {
		++(shapedata.polygon.polygon->referencecount);
	}
#endif
	
	++world->shapecount;
	++body->shapecount;
	
}

void ep_Shape::DeInit() {
	
	// destroy contacts
	while(first_contactlink!=NULL) world->DestroyContact(first_contactlink->contact);
	
	// remove from linked list
	if(prev==NULL) body->first_shape = next;
	else prev->next = next;
	if(next==NULL) body->last_shape = prev;
	else next->prev = prev;
	if(body->current_shape==this) body->current_shape = NULL;
	
#if EP_COMPILE_DEBUGCHECKS
	if(shapetype==EP_SHAPETYPE_POLYGON) {
		--(shapedata.polygon.polygon->referencecount);
	}
#endif
	
	--world->shapecount;
	--body->shapecount;
	
}

void ep_Shape::UpdateAABB() {
	// calculates the AABB and some other vars
	
	if(updateaabb) {
		
		// calculate world coordinates of shape origin
		xxx = body->x+ep_transform_x(body->rot_sin, body->rot_cos, x-body->xoff, y-body->yoff);
		yyy = body->y+ep_transform_y(body->rot_sin, body->rot_cos, x-body->xoff, y-body->yoff);
		
		// calculate total transformation
		sin_t = ep_totaltransform_sin(body->rot_sin, body->rot_cos, rot_sin, rot_cos);
		cos_t = ep_totaltransform_cos(body->rot_sin, body->rot_cos, rot_sin, rot_cos);
		
		// set up AABB (Axis Aligned Bounding Box)
		if(shapetype==EP_SHAPETYPE_BOX) {
			double xx = ep_abs(cos_t)*shapedata.box.w+ep_abs(sin_t)*shapedata.box.h,
			       yy = ep_abs(sin_t)*shapedata.box.w+ep_abs(cos_t)*shapedata.box.h;
			aabb_x1 = xxx-xx;
			aabb_y1 = yyy-yy;
			aabb_x2 = xxx+xx;
			aabb_y2 = yyy+yy;
		}
		if(shapetype==EP_SHAPETYPE_CIRCLE) {
			aabb_x1 = xxx-shapedata.circle.r;
			aabb_y1 = yyy-shapedata.circle.r;
			aabb_x2 = xxx+shapedata.circle.r;
			aabb_y2 = yyy+shapedata.circle.r;
		}
		if(shapetype==EP_SHAPETYPE_POLYGON) {
			ep_PolygonVertex *vertex = shapedata.polygon.polygon->vertices;
			unsigned long vc = shapedata.polygon.polygon->vertexcount;
			double xx = xxx+ep_transform_x(sin_t, cos_t, vertex[0].x, vertex[0].y),
			       yy = yyy+ep_transform_y(sin_t, cos_t, vertex[0].x, vertex[0].y);
			aabb_x1 = xx;
			aabb_y1 = yy;
			aabb_x2 = xx;
			aabb_y2 = yy;
			for(unsigned long i = 1; i<vc; ++i) {
				xx = xxx+ep_transform_x(sin_t, cos_t, vertex[i].x, vertex[i].y);
				yy = yyy+ep_transform_y(sin_t, cos_t, vertex[i].x, vertex[i].y);
				if(xx<aabb_x1) aabb_x1 = xx;
				if(yy<aabb_y1) aabb_y1 = yy;
				if(xx>aabb_x2) aabb_x2 = xx;
				if(yy>aabb_y2) aabb_y2 = yy;
			}
		}
		
		updateaabb = false;
	}
	
}

// contacts

ep_Contact* ep_Shape::GetFirstContact() {
	if(first_contactlink==NULL) return NULL;
	else return first_contactlink->contact;
}

ep_Contact* ep_Shape::GetLastContact() {
	if(last_contactlink==NULL) return NULL;
	else return last_contactlink->contact;
}

ep_Contact* ep_Shape::GetPreviousContact(ep_Contact* contact) {
	if(this==contact->shape1) {
		if(contact->link1.prev==NULL) return NULL;
		else return contact->link1.prev->contact;
	}
	if(this==contact->shape2) {
		if(contact->link2.prev==NULL) return NULL;
		else return contact->link2.prev->contact;
	}
	return NULL;
}

ep_Contact* ep_Shape::GetNextContact(ep_Contact* contact) {
	if(this==contact->shape1) {
		if(contact->link1.next==NULL) return NULL;
		else return contact->link1.next->contact;
	}
	if(this==contact->shape2) {
		if(contact->link2.next==NULL) return NULL;
		else return contact->link2.next->contact;
	}
	return NULL;
}

// other

void ep_Shape::SetMaterial(float _restitution, float _friction, double _normalvelocity, double _tangentvelocity) {
	if(restitution!=_restitution || friction!=_friction || normalvelocity!=_normalvelocity || _tangentvelocity!=_tangentvelocity) {
		restitution = _restitution;
		friction = _friction;
		normalvelocity = _normalvelocity;
		tangentvelocity = _tangentvelocity;
		body->Awake(false, true);
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed material of shape %lu of body %lu in world %lu.", id, body->id, world->id);
#endif
}

void ep_Shape::SetCollision(unsigned long _collidemask1, unsigned long _collidemask2, unsigned long _group) {
	if(collidemask1!=_collidemask1 || collidemask2!=_collidemask2 || group!=_group) {
		collidemask1 = _collidemask1;
		collidemask2 = _collidemask2;
		group = _group;
		updatecontacts = true;
		// Awake is not needed
	}
#if EP_COMPILE_DEBUGMESSAGES
	main->Message(EP_MESSAGELEVEL_NORMALCHANGE, "Changed collision settings of shape %lu of body %lu in world %lu.", id, body->id, world->id);
#endif
}

