/*
 * Copyright 2009-2010 Maarten Baert
 * maarten-baert@hotmail.com
 * http://www.maartenbaert.be/
 * 
 * This file is part of ExtremePhysics.
 * 
 * ExtremePhysics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * ExtremePhysics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ExtremePhysics. If not, see <http://www.gnu.org/licenses/>.
 * 
 * File: gm_world.cpp
 * Wrapper for ep_World.
 */

#include "gm.h"

gmexport double ep_world_create() {
	ep_World *world;
	if((world = epmain.CreateWorld(sizeof(gmuserdata)))==NULL) {
		return 0;
	}
	for(int i = 0; i<GM_USERVARS; ++i) {
		((gmuserdata*)(world->GetUserData()))->var[i] = 0;
	}
	return world->GetID();
}

gmexport double ep_world_destroy(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_destroy: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (epmain.DestroyWorld(world))? 1 : 0;
}

gmexport double ep_world_exists(double world_id) {
	return (epmain.FindWorld(gm_cast<unsigned long>(world_id))!=NULL)? 1 : 0;
}

gmexport double ep_world_clear(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_clear: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	world->Clear();
	return 1;
}

gmexport double ep_world_set_settings(double world_id, double timestep, double iterations, double contact_threshold, double velocity_threshold, double baumgarte_factor, double mass_bias) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_set_settings: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	world->SetSettings(timestep, gm_cast<unsigned long>(iterations), contact_threshold, velocity_threshold, baumgarte_factor, mass_bias);
	return 1;
}

gmexport double ep_world_set_primary_axis(double world_id, double horizontal) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_set_primary_axis: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	world->SetPrimaryAxis(gm_cast<bool>(horizontal));
	return 1;
}

gmexport double ep_world_set_sleeping(double world_id, double enable_sleeping, double time_stable, double time_outofview, double stable_maxvel, double stable_maxrotvel) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_set_sleeping: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	world->SetSleeping(gm_cast<bool>(enable_sleeping), time_stable, time_outofview, stable_maxvel, stable_maxrotvel);
	return 1;
}

gmexport double ep_world_simulate_step(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_simulate_step: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->SimulateStep())? 1 : 0;
}

gmexport double ep_world_previous(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_previous: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetPrevious()==NULL)? 0 : world->GetPrevious()->GetID();
}

gmexport double ep_world_next(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_next: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetNext()==NULL)? 0 : world->GetNext()->GetID();
}

gmexport double ep_world_set_uservar(double world_id, double index, double value) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_set_uservar: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	int i = gm_cast<int>(index);
	if(i<0 || i>=GM_USERVARS) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "Can't set user variable of world %lu, index is out of range.", world->GetID());
		return 0;
	}
	((gmuserdata*)(world->GetUserData()))->var[i] = value;
	return 1;
}

gmexport double ep_world_get_uservar(double world_id, double body_id, double index) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_get_uservar: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	int i = gm_cast<int>(index);
	if(i<0 || i>=GM_USERVARS) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "Can't get user variable of world %lu, index is out of range.", world->GetID());
		return 0;
	}
	return ((gmuserdata*)(world->GetUserData()))->var[i];
}

gmexport double ep_world_first_polygon(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_first_polygon: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetFirstPolygon()==NULL)? 0 : world->GetFirstPolygon()->GetID();
}

gmexport double ep_world_last_polygon(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_last_polygon: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetLastPolygon()==NULL)? 0 : world->GetLastPolygon()->GetID();
}

gmexport double ep_world_polygon_count(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_polygon_count: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return world->GetPolygonCount();
}

gmexport double ep_world_first_body(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_first_body: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetFirstBody()==NULL)? 0 : world->GetFirstBody()->GetID();
}

gmexport double ep_world_last_body(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_last_body: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetLastBody()==NULL)? 0 : world->GetLastBody()->GetID();
}

gmexport double ep_world_body_count(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_body_count: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return world->GetBodyCount();
}

gmexport double ep_world_first_contact(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_first_contact: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetFirstContact()==NULL)? 0 : world->GetFirstContact()->GetID();
}

gmexport double ep_world_last_contact(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_last_contact: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetLastContact()==NULL)? 0 : world->GetLastContact()->GetID();
}

gmexport double ep_world_contact_count(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_contact_count: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return world->GetContactCount();
}

gmexport double ep_world_first_hingejoint(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_first_hingejoint: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetFirstHingeJoint()==NULL)? 0 : world->GetFirstHingeJoint()->GetID();
}

gmexport double ep_world_last_hingejoint(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_last_hingejoint: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetLastHingeJoint()==NULL)? 0 : world->GetLastHingeJoint()->GetID();
}

gmexport double ep_world_hingejoint_count(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_hingejoint_count: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return world->GetHingeJointCount();
}

gmexport double ep_world_first_view(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_first_view: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetFirstView()==NULL)? 0 : world->GetFirstView()->GetID();
}

gmexport double ep_world_last_view(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_last_view: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return (world->GetLastView()==NULL)? 0 : world->GetLastView()->GetID();
}

gmexport double ep_world_view_count(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_view_count: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return world->GetViewCount();
}

gmexport double ep_world_shape_count(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_shape_count: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return world->GetShapeCount();
}

gmexport double ep_world_force_count(double world_id) {
	ep_World *world;
	if((world = epmain.FindWorld(gm_cast<unsigned long>(world_id)))==NULL) {
		epmain.Message(EP_MESSAGELEVEL_ERROR, "ep_world_force_count: World %lu doesn't exist.", gm_cast<unsigned long>(world_id));
		return 0;
	}
	return world->GetForceCount();
}

